@@ -8,7 +8,7 @@
  *
  *
  * IDENTIFICATION
- *	  $Id: pgtclCmds.c,v 1.1 2004/02/29 20:54:55 lbayuk Exp $
+ *	  $Id: pgtclCmds.c,v 1.5 2004/06/13 20:44:59 lbayuk Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -275,12 +275,226 @@ Pg_disconnect(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST obj
 	return Tcl_UnregisterChannel(interp, conn_chan);
 }
 
+/**********************************
+ * get_result_format - Helper for pg_exec_prepared and pg_exec_params
+
+	Parse resultListObj and make resultFormat argument.
+	The Tcl command syntax supports per-column result formats, but libpq
+	does not (yet), so make sure the caller isn't asking for something
+	that libpq can't handle. Take the first value, and make sure the
+	remaining ones match.
+	On success, store the result format in *resultFormat and return TCL_OK.
+	On error, store a message in the interp and return TCL_ERROR.
+
+ **********************************/
+static int
+get_result_format(Tcl_Interp *interp, Tcl_Obj *resultListObj, int *resultFormat)
+{
+	int		   listLen;
+	Tcl_Obj	   **objp;
+	int		i;
+
+	if (Tcl_ListObjGetElements(interp, resultListObj, &listLen, &objp) != TCL_OK)
+	{
+		Tcl_SetResult(interp, "Invalid resultFormatList parameter", TCL_STATIC);
+		return TCL_ERROR;
+	}
+	if (listLen > 0)
+	{
+		*resultFormat = (*Tcl_GetString(objp[0]) == 'B');
+		for (i = 1; i < listLen; i++)
+			if (*resultFormat != (*Tcl_GetString(objp[i]) == 'B'))
+			{
+				Tcl_SetResult(interp, "Mixed resultFormat values are not supported",
+					TCL_STATIC);
+				return TCL_ERROR;
+			}
+	}
+    else *resultFormat = 0;  /* Empty list => All TEXT */
+
+	return TCL_OK;;
+}
+
+/**********************************
+ * get_param_formats - Helper for pg_exec_prepared and pg_exec_params
+
+	Parse argFormat list and make paramFormats argument.
+	The parameter must either be empty (all TEXT), a single word T*|B*
+	(all that format), or nParams values, one per query parameter.
+    The libpq call will accept a null pointer for this argument, meaning
+	all text, so we can avoid the allocation in that most common case.
+	Set the allParamsText flag in that case.
+
+	On success, set or clear the allParamsText flag, store a NULL
+	pointer or a pointer to nParams ints in paramFormatsResult,
+	and return TCL_OK. (Caller must free paramFormatsResult.)
+	On error, store a message in the interp and return TCL_ERROR.
+
+ **********************************/
+static int
+get_param_formats(Tcl_Interp *interp, Tcl_Obj *argFormatListObj,
+	int nParams, int *allParamsText, int **paramFormatsResult)
+{
+	int		listLen;
+	Tcl_Obj	**objp;
+	int		*paramFormats;
+	int		i;
+
+	if (Tcl_ListObjGetElements(interp, argFormatListObj, &listLen, &objp) != TCL_OK)
+	{
+		Tcl_SetResult(interp, "Invalid argFormatList parameter", TCL_STATIC);
+		return TCL_ERROR;
+	}
+
+	paramFormats = NULL;
+	*allParamsText = 1;
+	if (listLen == 1)
+	{
+		if (*Tcl_GetString(objp[0]) == 'B')
+		{
+			paramFormats = (int *)Tcl_Alloc(nParams * sizeof(int));
+			for (i = 0; i < nParams; i++)
+				paramFormats[i] = 1;
+			*allParamsText = 0;
+		}
+	}
+	else if (listLen > 1)
+	{
+		if (listLen != nParams)
+		{
+			Tcl_SetResult(interp, "Mismatched argFormatList and parameter count",
+				TCL_STATIC);
+			return TCL_ERROR;
+		}
+		paramFormats = (int *)Tcl_Alloc(nParams * sizeof(int));
+		for (i = 0; i < nParams; i++)
+			if ((paramFormats[i] = (*Tcl_GetString(objp[i]) == 'B')))
+				*allParamsText = 0;
+	}
+
+	*paramFormatsResult = paramFormats;
+	return TCL_OK;
+}
+
+/**********************************
+ * get_param_values - Helper for pg_exec, pg_exec_prepared, and pg_exec_params
+
+	For each query parameter, we need its address in an array paramValues.
+	For each binary-format query parameter, we need its length in an
+	array paramLengths.  (Length is ignored for text-format parameters.)
+	If there are no binary parameters, paramLengths will be NULL.
+	(If this is known in advance, and the allParamText flag is 1, then the
+	the paramLengths_result argument can be supplied as NULL. This is used
+	by the extended form of pg_exec.)
+	If there are no query parameters, both arrays are NULL.
+	CHECK: Currently uses ByteArray for binary, String for text, but it
+	is unclear if this is correct.
+	 
+	Stores the results in *paramLengths_result and *paramValues_result,
+	which the caller must free if not NULL.
+	No errors, void return.
+
+ **********************************/
+static void
+get_param_values(Tcl_Interp *interp, Tcl_Obj *CONST *objv,
+	int nParams, int allParamsText, int *paramFormats,
+	const char *const **paramValues_result, int **paramLengths_result)
+{
+	int		i;
+	int		*paramLengths;
+	const char	**paramValues;
+
+	paramLengths = NULL;
+	paramValues = NULL;
+	if (nParams > 0)
+	{
+		paramValues = (const char **)Tcl_Alloc(nParams * sizeof(char *));
+		if (!allParamsText)
+			paramLengths = (int *)Tcl_Alloc(nParams * sizeof(int));
+
+		for (i = 0; i < nParams; i++)
+		{
+			if (paramFormats && paramFormats[i]) /* Binary Format */
+				paramValues[i] = (char *)Tcl_GetByteArrayFromObj(*objv,
+										&paramLengths[i]);
+			else /* Text Format */
+				paramValues[i] = Tcl_GetString(*objv);
+			objv++;
+		}
+	}
+	*paramValues_result = paramValues;
+	if (paramLengths_result)
+		*paramLengths_result = paramLengths;
+}
+
+/**********************************
+ * get_param_types - Helper for pg_exec_params
+
+	Build an array of type OIDs from the supplied list. The list must
+	either be empty or contain nParams items.
+	 
+	Stores the result in *paramTypes, which the caller must free
+	if not NULL. This will be either NULL or a pointer to nParams Oids.
+	Returns TCL_OK if OK.
+	On error, store a message in the interp and return TCL_ERROR.
+
+ **********************************/
+static int
+get_param_types(Tcl_Interp *interp, Tcl_Obj *argTypeListObj,
+	int nParams, Oid **paramTypesResult)
+{
+	int		listLen;
+	Tcl_Obj	**objp;
+	Oid *paramTypes;
+	int		i;
+
+	if (Tcl_ListObjGetElements(interp, argTypeListObj, &listLen, &objp) != TCL_OK)
+	{
+		Tcl_SetResult(interp, "Invalid argTypeList parameter", TCL_STATIC);
+		return TCL_ERROR;
+	}
+
+	paramTypes = NULL;
+	if (listLen > 0)
+	{
+		if (listLen != nParams)
+		{
+			Tcl_SetResult(interp, "Mismatched argTypeList and parameter count",
+				TCL_STATIC);
+			return TCL_ERROR;
+		}
+		paramTypes = (Oid *)Tcl_Alloc(nParams * sizeof(int));
+		for (i = 0; i < nParams; i++)
+		{
+			/*
+			 * Note: paramTypes[i] is Oid which is unsigned int, and
+			 * Tcl_GetIntFromObj() expects a pointer to a signed int.
+			 * There is no direct support for unsigned in Tcl, but tests
+			 * and code examination show it will work for values that
+			 * will fit in unsigned but not signed. Anyway, it's the best
+			 * we can do.
+			 */
+			if (Tcl_GetIntFromObj(interp, objp[i], (int *)&paramTypes[i]) != TCL_OK)
+			{
+		  		Tcl_Free((char *)paramTypes);
+				return TCL_ERROR;
+			}
+		}
+	}
+	*paramTypesResult = paramTypes;
+	return TCL_OK;
+}
+
 /**********************************
  * pg_exec
  send a query string to the backend connection
 
  syntax:
- pg_exec connection query
+ pg_exec connection query ?arg...?
+
+ Optional args are used as parameters to PQexecParams(). This is a simplified
+ version of Pg_exec_params using text-only, untyped parameters.
+ With no optional args, use regular PQexec().
 
  the return result is either an error message or a handle for a query
  result.  Handles start with the prefix "pgp"
@@ -294,10 +508,13 @@ Pg_exec(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
 	PGresult   *result;
 	char	   *connString;
 	char	   *execString;
+	const char *const *paramValues;
+	int		   nParams;
 
-	if (objc != 3)
+	nParams = objc - 3;
+	if (nParams < 0)
 	{
-		Tcl_WrongNumArgs(interp, 1, objv, "connection queryString");
+		Tcl_WrongNumArgs(interp, 1, objv, "connection queryString ?param...?");
 		return TCL_ERROR;
 	}
 
@@ -314,7 +531,23 @@ Pg_exec(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
 	}
 
 	execString = Tcl_GetString(objv[2]);
-	result = PQexec(conn, execString);
+
+	if (nParams > 0)
+	{
+		get_param_values(interp, &objv[3], nParams, /* allParamsText = */ 1,
+			/* paramFormats = */ NULL, &paramValues,
+			/* paramLengths_result = */ NULL);
+
+		result = PQexecParams(conn, execString, nParams, NULL,
+						paramValues, NULL, NULL, /* resultFormat= */ 0);
+
+		if (paramValues)
+			Tcl_Free((char *)paramValues);
+
+	} else {
+		result = PQexec(conn, execString);
+	}
+
 
 	/* Transfer any notify events from libpq to Tcl event queue. */
 	PgNotifyTransferEvents(connid);
@@ -354,6 +587,7 @@ Pg_exec(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
 }
 
 
+
 /**********************************
  * pg_exec_prepared
  Execute a pre-prepared query with supplied parameters
@@ -382,18 +616,15 @@ Pg_exec_prepared(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST
 {
 	Pg_ConnectionId *connid;
 	PGconn	   *conn;
-	PGresult   *result;
 	char	   *connString;
+	PGresult   *result;
 	char	   *statementName;
 	int		   nParams;
 	int		   allParamsText;
 	int 	   resultFormat;
-	int		   listLen;
-	Tcl_Obj	   **objp;
-	int		   i;
 	int		   *paramFormats;
 	int		   *paramLengths;
-	CONST char **paramValues;
+	const char *const *paramValues;
 	int		   returnValue;
 
 	nParams = objc - 5;
@@ -418,97 +649,20 @@ Pg_exec_prepared(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST
 
 	statementName = Tcl_GetString(objv[2]);
 
-	/*
-	 * Parse resultFormatList and make resultFormat argument.
-	 * The Tcl command syntax supports per-column result formats, but libpq
-	 * does not (yet), so make sure the caller isn't asking for something
-	 * that libpq can't handle. Take the first value, and make sure the
-	 * remaining ones match.
-	 */
-	if (Tcl_ListObjGetElements(interp, objv[3], &listLen, &objp) != TCL_OK)
-	{
-		Tcl_SetResult(interp, "Invalid resultFormatList parameter to pg_exec_prepared", TCL_STATIC);
+	/* Parse resultFormatList and make resultFormat argument. */
+	if (get_result_format(interp, objv[3], &resultFormat) != TCL_OK)
 		return TCL_ERROR;
-	}
-	if (listLen > 0)
-	{
-		resultFormat = (*Tcl_GetString(objp[0]) == 'B');
-		for (i = 1; i < listLen; i++)
-			if (resultFormat != (*Tcl_GetString(objp[i]) == 'B'))
-			{
-				Tcl_SetResult(interp, "Mixed resultFormat values are not "
-					"supported by pg_exec_prepared", TCL_STATIC);
-				return TCL_ERROR;
-			}
-	}
-    else resultFormat = 0;  /* Empty list => All TEXT */
 
-	/*
-	 * Parse argFormat list and make paramFormats argument.
-	 * The parameter must either be empty (all TEXT), a single word T*|B*
-	 * (all that format), or nParams values, one per query parameter.
-     * The libpq call will accept a null pointer for this argument, meaning
-	 * all text, so we can avoid the allocation in that most common case.
-	 */
-	if (Tcl_ListObjGetElements(interp, objv[4], &listLen, &objp) != TCL_OK)
-	{
-		Tcl_SetResult(interp, "Invalid argFormatList parameter to pg_exec_prepared", TCL_STATIC);
+	/* Parse argFormat list and make paramFormats argument and all-text flag */
+	if (get_param_formats(interp, objv[4], nParams, &allParamsText,
+			&paramFormats) != TCL_OK)
 		return TCL_ERROR;
-	}
 
-	paramFormats = NULL;
-	allParamsText = 1;
-	if (listLen == 1)
-	{
-		if (*Tcl_GetString(objp[0]) == 'B')
-		{
-			paramFormats = (int *)Tcl_Alloc(nParams * sizeof(int));
-			for (i = 0; i < nParams; i++)
-				paramFormats[i] = 1;
-			allParamsText = 0;
-		}
-	}
-	else if (listLen > 1)
-	{
-		if (listLen != nParams)
-		{
-			Tcl_SetResult(interp, "Mismatched argFormatList and parameter "
-				"count in pg_exec_prepared", TCL_STATIC);
-			return TCL_ERROR;
-		}
-		paramFormats = (int *)Tcl_Alloc(nParams * sizeof(int));
-		for (i = 0; i < nParams; i++)
-			if ((paramFormats[i] = (*Tcl_GetString(objp[i]) == 'B')))
-				allParamsText = 0;
-	}
-
-	/*
-	 * For each query parameter, we need its address in an array paramValues.
-	 * For each binary-format query parameter, we need its length in an
-	 * array paramLengths.  (Length is ignored for text-format parameters.)
-	 * If there are no binary parameters, paramLengths can be NULL.
-	 * If there are no query parameters, both arrays are NULL.
-	 * CHECK: Currently uses ByteArray for binary, String for text, but it
-	 * is unclear if this is correct.
-	 */
-	paramLengths = NULL;
-	paramValues = NULL;
-	if (nParams > 0)
-	{
-		paramValues = (CONST char **)Tcl_Alloc(nParams * sizeof(char *));
-		if (!allParamsText)
-			paramLengths = (int *)Tcl_Alloc(nParams * sizeof(int));
-
-		for (i = 0; i < nParams; i++)
-		{
-			if (paramFormats && paramFormats[i]) /* Binary Format */
-				paramValues[i] = (char *)Tcl_GetByteArrayFromObj(objv[i+5],
-										&paramLengths[i]);
-			else /* Text Format */
-				paramValues[i] = Tcl_GetString(objv[i+5]);
-		}
-	}
+	/* Copy query parameters, and lengths if binary format */
+	get_param_values(interp, &objv[5], nParams, allParamsText, paramFormats,
+			&paramValues, &paramLengths);
 
+	/* Now execute the prepared query */
 	result = PQexecPrepared(conn, statementName, nParams, paramValues,
 			paramLengths, paramFormats, resultFormat);
 
@@ -544,6 +698,135 @@ Pg_exec_prepared(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST
 	return returnValue;
 }
 
+/**********************************
+ * pg_exec_params
+ Parse, bind parameters, and execute a query
+
+ Syntax:
+ pg_exec_params connection query resultFormatList argFormatList
+         artTypeList arg...
+
+ query is an SQL statement with parameter placeholders specified as
+ $1, $2, etc.
+
+ argFormatList is empty (= same as T), a single word T|B|TEXT|BINARY, or
+ a list of those words, describing each argument as text or binary. If a
+ single word, it applies to all arguments.  (Actually, anything starting
+ with B means Binary, and anything else means Text. There is no error
+ checking.)
+
+ resultFormatList is similar to argFormatList except that it applies to the
+ columns of the results. Currently,  all result parameters must be text, or
+ all must be binary (this is a libpq limitation, not a PostgreSQL
+ limitation). So you might as well specify a single word BINARY or leave it
+ empty.
+
+ argTypeList is a list of PostgreSQL type OIDs for the query parameter
+ arguments. Type OIDs must be supplied for each binary-format argument.
+ If there are any binary format arguments, the argTypeList must contain
+ an entry for each argument, although the actual value will be ignored
+ for text-mode arguments.
+
+ The return result is either an error message or a handle for a query
+ result.
+ **********************************/
+
+int
+Pg_exec_params(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
+{
+	Pg_ConnectionId *connid;
+	PGconn	   *conn;
+	char	   *connString;
+	PGresult   *result;
+	char	   *queryString;
+	int		   nParams;
+	int		   allParamsText;
+	int 	   resultFormat;
+	int		   *paramFormats;
+	int		   *paramLengths;
+	const char *const *paramValues;
+    Oid		   *paramTypes;
+	int		   returnValue;
+
+	nParams = objc - 6;
+	if (nParams < 0)
+	{
+		Tcl_WrongNumArgs(interp, 1, objv, "connection queryString "
+			"resultFormat argFormatList argTypeList arg...");
+		return TCL_ERROR;
+	}
+
+	connString = Tcl_GetString(objv[1]);
+
+	conn = PgGetConnectionId(interp, connString, &connid);
+	if (conn == (PGconn *)NULL)
+		return TCL_ERROR;
+
+	if (connid->res_copyStatus != RES_COPY_NONE)
+	{
+		Tcl_SetResult(interp, "Attempt to query while COPY in progress", TCL_STATIC);
+		return TCL_ERROR;
+	}
+
+	queryString = Tcl_GetString(objv[2]);
+
+	/* Parse resultFormatList and make resultFormat argument. */
+	if (get_result_format(interp, objv[3], &resultFormat) != TCL_OK)
+		return TCL_ERROR;
+
+	/* Parse argFormat list and make paramFormats argument and all-text flag */
+	if (get_param_formats(interp, objv[4], nParams, &allParamsText,
+			&paramFormats) != TCL_OK)
+		return TCL_ERROR;
+
+	/* Get the parameter type OID list into an array */
+	if (get_param_types(interp, objv[5], nParams, &paramTypes) != TCL_OK) {
+		if (paramFormats)
+			Tcl_Free((char *)paramFormats);
+		return TCL_ERROR;
+	}
+
+	/* Copy query parameters, and lengths if binary format */
+	get_param_values(interp, &objv[6], nParams, allParamsText, paramFormats,
+			&paramValues, &paramLengths);
+
+	/* Now execute the parameterized query */
+	result = PQexecParams(conn, queryString, nParams, paramTypes,
+			paramValues, paramLengths, paramFormats, resultFormat);
+
+	/* Transfer any notify events from libpq to Tcl event queue. */
+	PgNotifyTransferEvents(connid);
+
+	/*
+	 * Note: You can't use this command to start a COPY, so there is no
+	 * need to check for PGRES_COPY_* status like pg_exec does.
+	 */
+	if (!result)
+	{
+		/* error occurred during the query */
+		Tcl_SetObjResult(interp, Tcl_NewStringObj(PQerrorMessage(conn), -1));
+		returnValue = TCL_ERROR;
+	}
+	else if (PgSetResultId(interp, connString, result) == -1)
+	{
+		/* Query response was OK, but unable to allocate result slot. */
+		PQclear(result);
+		returnValue = TCL_ERROR;
+	}
+	else
+		returnValue = TCL_OK;
+
+	if (paramFormats)
+		Tcl_Free((char *)paramFormats);
+	if (paramLengths)
+		Tcl_Free((char *)paramLengths);
+	if (paramValues)
+		Tcl_Free((char *)paramValues);
+	if (paramTypes)
+		Tcl_Free((char *)paramTypes);
+
+	return returnValue;
+}
 
 /**********************************
  * pg_result_errorfield_code
@@ -613,6 +896,8 @@ pg_result_errorfield_code(char *fieldName)
 
 	-cmdTuples	Same as -numTuples, but for DELETE and UPDATE
 
+	-cmdStatus	returns the command status tag, e.g. "INSERT ... ..."
+
 	-numAttrs	returns the number of attributes returned by the query
 
 	-assign arrayName
@@ -669,6 +954,7 @@ Pg_result(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
 		"-numTuples", "-cmdTuples", "-numAttrs", "-assign", "-assignbyidx",
 		"-getTuple", "-tupleArray", "-attributes", "-lAttributes",
 		"-lxAttributes", "-clear", "-list", "-llist", "-getNull",
+		"-cmdStatus",
 		(char *)NULL
 	};
 
@@ -677,7 +963,8 @@ Pg_result(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
 		OPT_STATUS, OPT_ERROR, OPT_ERRORFIELD, OPT_CONN, OPT_OID,
 		OPT_NUMTUPLES, OPT_CMDTUPLES, OPT_NUMATTRS, OPT_ASSIGN, OPT_ASSIGNBYIDX,
 		OPT_GETTUPLE, OPT_TUPLEARRAY, OPT_ATTRIBUTES, OPT_LATTRIBUTES,
-		OPT_LXATTRIBUTES, OPT_CLEAR, OPT_LIST, OPT_LLIST, OPT_GETNULL
+		OPT_LXATTRIBUTES, OPT_CLEAR, OPT_LIST, OPT_LLIST, OPT_GETNULL,
+		OPT_CMDSTATUS
 	};
 
 	/*
@@ -806,6 +1093,7 @@ Pg_result(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
 				Tcl_SetIntObj(Tcl_GetObjResult(interp), PQntuples(result));
 				return TCL_OK;
 			}
+
 		case OPT_CMDTUPLES:
 			{
 				if (objc != 3)
@@ -818,6 +1106,19 @@ Pg_result(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
 				return TCL_OK;
 			}
 
+		case OPT_CMDSTATUS:
+			{
+				if (objc != 3)
+				{
+					Tcl_WrongNumArgs(interp, 3, objv, NULL);
+					return TCL_ERROR;
+				}
+
+				Tcl_SetStringObj(Tcl_GetObjResult(interp), PQcmdStatus(result), -1);
+				return TCL_OK;
+			}
+
+
 		case OPT_NUMATTRS:
 			{
 				if (objc != 3)
@@ -1343,7 +1644,7 @@ Pg_execute(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]
 	int			i;
 	int			tupno;
 	int			ntup;
-	int			loop_rc;
+	int			loop_rc = TCL_OK;
 	char	   *array_varname = NULL;
 	char	   *arg;
 	char	   *connString;
@@ -1550,36 +1851,34 @@ Pg_execute(ClientData cData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]
 		loop_rc = Tcl_EvalObjEx(interp, evalObj, 0);
 
 		/* The returncode of the loop body controls the loop execution */
-		if (loop_rc == TCL_OK || loop_rc == TCL_CONTINUE)
+		if (loop_rc != TCL_OK && loop_rc != TCL_CONTINUE)
 		{
-			/* OK or CONTINUE means start next loop invocation */
-			continue;
-		}
+			if (loop_rc == TCL_ERROR)
+			{
+				/* Show where the error occurred */
+				char		msg[60];
 
-		if (loop_rc == TCL_RETURN)
-		{
-			/* RETURN means hand up the given interpreter result */
-			PQclear(result);
-			return TCL_RETURN;
-		}
+				sprintf(msg, "\n    (\"pg_execute\" body line %d)",
+						interp->errorLine);
+				Tcl_AddErrorInfo(interp, msg);
 
-		if (loop_rc == TCL_BREAK)
-		{
-			/* BREAK means leave the loop */
+			} else if (loop_rc == TCL_BREAK) {
+				/* On break, break out but return OK */
+				loop_rc = TCL_OK;
+			}
 			break;
 		}
-
-		PQclear(result);
-		return TCL_ERROR;
 	}
 
 	/*
 	 * At the end of the loop we put the number of rows we got into the
-	 * interpreter result and clear the result set.
+	 * interpreter result, but only on normal return, and clear the result set.
 	 */
-	Tcl_SetObjResult(interp, Tcl_NewIntObj(ntup));
+	if (loop_rc == TCL_OK)
+		Tcl_SetObjResult(interp, Tcl_NewIntObj(ntup));
 	PQclear(result);
-	return TCL_OK;
+
+	return loop_rc;
 }
 
 
@@ -1596,6 +1895,10 @@ execute_put_values(Tcl_Interp *interp, char *array_varname,
 	int			i;
 	int			n;
 	Tcl_Obj	   *value;
+	/*
+	 * Note: "gcc -Wall" reports that the following two variables
+	 * "may be used uninitialized", but that is not true.
+	 */
 	char	   *varname;
 	char	   *indexname;
 
@@ -2909,6 +3212,7 @@ Pg_on_connection_loss(ClientData cData, Tcl_Interp *interp, int objc,
 /***********************************
 Pg_escape_string
 	escape string for inclusion in SQL queries
+    See also Pg_quote
 
  syntax:
    pg_escape_string string
@@ -2922,7 +3226,6 @@ Pg_escape_string(ClientData cData, Tcl_Interp *interp, int objc,
 	char	   *toString;
 	int         fromStringLen;
 	int			toStringLen;
-	Tcl_Obj	   *resultObj;
 
 	if (objc != 2)
 	{
@@ -2952,6 +3255,61 @@ Pg_escape_string(ClientData cData, Tcl_Interp *interp, int objc,
 	return TCL_OK;
 }
 
+/***********************************
+Pg_quote
+	escape and quote string for inclusion in SQL queries
+    See also Pg_escape_string
+    Note: First there was pg_escape_string, based on libpq PQescapeString.
+    This was added to Gborg libpgtcl CVS after 1.4b3. About a year later,
+    pg_escape_string was removed and pg_quote was added instead. But both
+    pgintcl and pgtcl-ng already released a version with pg_escape_string.
+    I'm not about to break things, so both are supported.
+
+ syntax:
+   pg_quote string
+
+***********************************/
+int
+Pg_quote(ClientData cData, Tcl_Interp *interp, int objc,
+				 Tcl_Obj *CONST objv[])
+{
+	char	   *fromString;
+	char	   *toString;
+	int         fromStringLen;
+	int			toStringLen;
+
+	if (objc != 2)
+	{
+		Tcl_WrongNumArgs(interp, 1, objv, "string");
+		return TCL_ERROR;
+	}
+
+	/*
+	 * Get the "from" string.
+	 */
+	fromString = Tcl_GetStringFromObj(objv[1], &fromStringLen);
+
+	/* 
+	 * Allocate the "to" string. Max size is documented in the
+	 * PostgreSQL docs as 2 * fromStringLen + 1. Add 2 for quotes,
+	 * and subtract 1 because NewStringObj doesn't need the ending null.
+	 */
+	toString = (char *) ckalloc((2 * fromStringLen) + 2);
+
+	/*
+	 * Call the library routine to escape the string, and return
+	 * the command result as a Tcl object with quote marks around it.
+	 */
+	toString[0] = '\'';
+	toStringLen = 1 + PQescapeString (toString+1, fromString, fromStringLen);
+	toString[toStringLen++] = '\'';
+	Tcl_SetObjResult(interp, Tcl_NewStringObj(toString, toStringLen));
+	ckfree(toString);
+
+	return TCL_OK;
+}
+
+
 /**********************************
  * pg_transaction_status
  Return the transaction status of a connection
